/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.registry.users.def;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.users.AccessControlAdmin;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.def.util.DefaultDatabaseUtil;

/**
 * @see org.wso2.usermanager.AccessControlAdmin
 */
public class DefaultAccessControlAdmin
        extends DefaultAuthorizer implements AccessControlAdmin {

    private static Log log = LogFactory.getLog(DefaultAccessControlAdmin.class);

    public DefaultAccessControlAdmin(DataSource dataSource, String algo) {
        super(dataSource, algo);
    }

    public DefaultAccessControlAdmin(DataSource dataSource, String algo,
            DefaultDataSupplier store) {
        super(dataSource, algo, store);
    }

    public DefaultAccessControlAdmin(DataSource dataSource,
            DefaultDataSupplier store) {
        super(dataSource, DefaultRealmConfig.PERMISSION_BLOCK_FIRST, store);
    }

    public void clearUserAuthorization(String userName, String resourceId,
            String action) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement getPermission = dbConnection
                    .prepareStatement(data
                            .getAuthorizerSQL(DefaultRealmConstants.GET_PERMISSION));
            getPermission.setString(1, resourceId);
            getPermission.setString(2, action);

            ResultSet rs = getPermission.executeQuery();
            int pid = -1;
            if (rs.next()) {
                pid = rs.getInt(data
                        .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));
            } else {
                throw new UserStoreException(Messages.getMessage("nullData"));
            }
            int uid = DefaultDatabaseUtil.getUserId(dbConnection, userName);

            PreparedStatement clearUPStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.DELETE_USER_PERMISSION));
            clearUPStmt.setInt(1, uid);
            clearUPStmt.setInt(2, pid);
            clearUPStmt.executeUpdate();
            dbConnection.commit();
            getPermission.close();
            clearUPStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void denyUser(String userName, String resourceId, String action)
            throws UserStoreException {
        short deny = 0;
        addAllowDenyUserPermission(userName, resourceId, action, deny);
    }

    public void authorizeRole(String roleName, String resourceId, String action)
            throws UserStoreException {
        short allow = 1;
        addAllowDenyRolePermission(roleName, resourceId, action, allow);
        
    }

    public void denyRole(String roleName, String resourceId, String action)
            throws UserStoreException {
        short deny = 0;
        addAllowDenyRolePermission(roleName, resourceId, action, deny);
    }

    public void clearRoleAuthorization(String roleName, String resourceId,
            String action) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement getPermission = dbConnection
                    .prepareStatement(data
                            .getAuthorizerSQL(DefaultRealmConstants.GET_PERMISSION));
            getPermission.setString(1, resourceId);
            getPermission.setString(2, action);

            ResultSet rs = getPermission.executeQuery();
            String pid = null;
            if (rs.next()) {
                pid = rs.getString(data
                        .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));
            } else {
                throw new UserStoreException(Messages.getMessage("nullData"));
            }

            int rid = DefaultDatabaseUtil.getRoleId(dbConnection, roleName);

            PreparedStatement deleteRolesStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.DELETE_ROLE_PERMISSION));
            deleteRolesStmt.setInt(1, rid);
            deleteRolesStmt.setString(2, pid);
            deleteRolesStmt.executeUpdate();

            dbConnection.commit();
            getPermission.close();
            deleteRolesStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }

    }

    public void clearResourceAuthorizations(String resourceId)
            throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement clearResourceAuthorizationsStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.DELETE_PERMISSION_ON_RESOURCE));
            clearResourceAuthorizationsStmt.setString(1, resourceId);
            clearResourceAuthorizationsStmt.executeUpdate();
            dbConnection.commit();
            clearResourceAuthorizationsStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void copyAuthorizations(String fromResourceId, String toResourceId)
            throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement getResourcePermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAuthorizerSQL(DefaultRealmConstants.GET_RESOURCE_PERMISSION));
            getResourcePermissionStmt.setString(1, toResourceId);
            ResultSet rs = getResourcePermissionStmt.executeQuery();
            if (rs.next()) {
                throw new UserStoreException(Messages
                        .getMessage("copyAuthorizationFailedPermissionExist"));
            }

            getResourcePermissionStmt.setString(1, fromResourceId);
            rs = getResourcePermissionStmt.executeQuery();
            PreparedStatement addPermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.ADD_PERMISSION));
            while (rs.next()) {
                String action = rs.getString(data
                        .getColumnName(DefaultRealmConstants.COLUMN_ID_ACTION));
                String oldPermissionId = rs.getString(data
                        .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));

                addPermissionStmt.setString(1, toResourceId);
                addPermissionStmt.setString(2, action);
                addPermissionStmt.executeUpdate();

                // ***** read the id
                PreparedStatement getPermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAuthorizerSQL(DefaultRealmConstants.GET_PERMISSION));
                getPermissionStmt.setString(1, toResourceId);
                getPermissionStmt.setString(2, action);

                ResultSet rs2 = getPermissionStmt.executeQuery();

                int idPermission = -1;
                if (rs2.next()) {
                    idPermission = rs2.getInt(data
                            .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));
                }
                // ***** end read id

                PreparedStatement getRolePermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAuthorizerSQL(DefaultRealmConstants.GET_ROLE_PERMISSION));

                getRolePermissionStmt.setString(1, oldPermissionId);
                ResultSet rsr = getRolePermissionStmt.executeQuery();
                PreparedStatement addRolePermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAccessControlAdminSQL(DefaultRealmConstants.ADD_ROLE_PERMISSION));

                while (rsr.next()) {

                    String roleId = rsr
                            .getString(data
                                    .getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_ID));
                    short allow = rsr
                            .getShort(data
                                    .getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED));

                    addRolePermissionStmt.setInt(1, idPermission);
                    addRolePermissionStmt.setShort(2, allow);
                    addRolePermissionStmt.setString(3, roleId);
                    addRolePermissionStmt.executeUpdate();
                }

                PreparedStatement getUserPermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAuthorizerSQL(DefaultRealmConstants.GET_USER_PERMISSION));

                getUserPermissionStmt.setString(1, oldPermissionId);
                ResultSet rsu = getUserPermissionStmt.executeQuery();

                PreparedStatement addUserPermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAccessControlAdminSQL(DefaultRealmConstants.ADD_USER_PERMISSION));

                while (rsu.next()) {
                    String userId = rsu
                            .getString(data
                                    .getColumnName(DefaultRealmConstants.COLUMN_ID_USER_ID));
                    short allow = rsu
                            .getShort(data
                                    .getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED));

                    addUserPermissionStmt.setInt(1, idPermission);
                    addUserPermissionStmt.setShort(2, allow);
                    addUserPermissionStmt.setString(3, userId);
                    addUserPermissionStmt.executeUpdate();
                }

                getRolePermissionStmt.close();
                addRolePermissionStmt.close();
                getUserPermissionStmt.close();
                addUserPermissionStmt.close();

            }

            getResourcePermissionStmt.close();
            addPermissionStmt.close();

            dbConnection.commit();

        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorCopyingAuthorizations"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }

    }

    public void authorizeUser(String userName, String resourceId, String action)
            throws UserStoreException {
        short allow = 1;
        addAllowDenyUserPermission(userName, resourceId, action, allow);
    }
    
    

    /**
     * Gets the permission if it exists, else create the permission and returns
     * the id.
     * 
     * @return
     */
    protected int getOrAddPermissionId(Connection dbConnection, String resourceId, String action)
            throws UserStoreException {
        int permissionId = -1;
        try {
            PreparedStatement getPermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAuthorizerSQL(DefaultRealmConstants.GET_PERMISSION));
            getPermissionStmt.setString(1, resourceId);
            getPermissionStmt.setString(2, action);

            ResultSet rs = getPermissionStmt.executeQuery();

            if (rs.next()) {
                permissionId = rs.getInt(data
                        .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));
            }

            PreparedStatement addPermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.ADD_PERMISSION));

            if (permissionId == -1) {
                // no permission so add it
                addPermissionStmt.setString(1, resourceId);
                addPermissionStmt.setString(2, action);
                addPermissionStmt.executeUpdate();
                dbConnection.commit();

                // now read the id again
                getPermissionStmt = dbConnection
                        .prepareStatement(data
                                .getAuthorizerSQL(DefaultRealmConstants.GET_PERMISSION));
                getPermissionStmt.setString(1, resourceId);
                getPermissionStmt.setString(2, action);

                ResultSet rs2 = getPermissionStmt.executeQuery();

                if (rs2.next()) {
                    permissionId = rs2.getInt(data
                            .getColumnName(DefaultRealmConstants.COLUMN_ID_ID));
                }

                dbConnection.commit();
            }

        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } 

        return permissionId;

    }
    
    
    protected  void addAllowDenyUserPermission(String userName, String resourceId, String action, short allow)
        throws UserStoreException{
        Connection dbConnection = null;

        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            
           
            int userId = DefaultDatabaseUtil.getUserId(dbConnection, userName);

            if (userId == -1) {
                throw new UserStoreException(Messages.getMessage("nullUser"));
            }

            int permissionId = this.getOrAddPermissionId(dbConnection, resourceId, action);
            
            PreparedStatement clearUPStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.DELETE_USER_PERMISSION));
            clearUPStmt.setInt(1, userId);
            clearUPStmt.setInt(2, permissionId);
            clearUPStmt.executeUpdate();

            PreparedStatement addUserPermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.ADD_USER_PERMISSION));
            addUserPermissionStmt.setInt(1, permissionId);
            addUserPermissionStmt.setShort(2, allow);
            addUserPermissionStmt.setInt(3, userId);
            addUserPermissionStmt.executeUpdate();
            dbConnection.commit();
            addUserPermissionStmt.close();
            clearUPStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }

    }

    protected  void addAllowDenyRolePermission(String roleName, String resourceId, String action, short allow)
                                        throws UserStoreException{
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages
                        .getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            int roleId = DefaultDatabaseUtil.getRoleId(dbConnection, roleName);
            
            if (roleId == -1) {
                throw new UserStoreException(Messages.getMessage("nullRole"));
            }

            int permissionId = this.getOrAddPermissionId(dbConnection, resourceId, action);
            
            PreparedStatement deleteRolesStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.DELETE_ROLE_PERMISSION));
            deleteRolesStmt.setInt(1, roleId);
            deleteRolesStmt.setInt(2, permissionId);
            deleteRolesStmt.executeUpdate();

            PreparedStatement addRolePermissionStmt = dbConnection
                    .prepareStatement(data
                            .getAccessControlAdminSQL(DefaultRealmConstants.ADD_ROLE_PERMISSION));
            addRolePermissionStmt.setInt(1, permissionId);
            addRolePermissionStmt.setShort(2, allow);
            addRolePermissionStmt.setInt(3, roleId);
            addRolePermissionStmt.executeUpdate();
            dbConnection.commit();
            addRolePermissionStmt.close();
            deleteRolesStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages
                    .getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages
                        .getMessage("errorClosingConnection"), e);
            }
        }
    }
    
}
